@@ -13,6 +13,7 @@
 #include <ctype.h>
 
 #include <libmount.h>
+#include <libsmartcols.h>
 
 #include "c.h"
 #include "nls.h"
@@ -25,7 +26,6 @@
 #include "swapheader.h"
 #include "swapon-common.h"
 #include "strutils.h"
-#include "tt.h"
 
 #define PATH_MKSWAP	"/sbin/mkswap"
 
@@ -34,9 +34,20 @@
 #endif
 
 #ifndef SWAP_FLAG_DISCARD
-# define SWAP_FLAG_DISCARD	0x10000 /* discard swap cluster after use */
+# define SWAP_FLAG_DISCARD	0x10000 /* enable discard for swap */
 #endif
 
+#ifndef SWAP_FLAG_DISCARD_ONCE
+# define SWAP_FLAG_DISCARD_ONCE 0x20000 /* discard swap area at swapon-time */
+#endif
+
+#ifndef SWAP_FLAG_DISCARD_PAGES
+# define SWAP_FLAG_DISCARD_PAGES 0x40000 /* discard page-clusters after use */
+#endif
+
+#define SWAP_FLAGS_DISCARD_VALID (SWAP_FLAG_DISCARD | SWAP_FLAG_DISCARD_ONCE | \
+				  SWAP_FLAG_DISCARD_PAGES)
+
 #ifndef SWAP_FLAG_PREFER
 # define SWAP_FLAG_PREFER	0x8000	/* set if swap priority specified */
 #endif
@@ -65,12 +76,9 @@ enum {
 	SIG_SWSUSPEND
 };
 
-#define SWAP_SIGNATURE		"SWAPSPACE2"
-#define SWAP_SIGNATURE_SZ	(sizeof(SWAP_SIGNATURE) - 1)
-
 static int all;
 static int priority = -1;	/* non-prioritized swap by default */
-static int discard;
+static int discard;		/* don't send swap discards by default */
 
 /* If true, don't complain if the device/file doesn't exist */
 static int ifexists;
@@ -81,19 +89,25 @@ static int verbose;
 struct colinfo {
         const char *name; /* header */
         double     whint; /* width hint (N < 1 is in percent of termwidth) */
-        int        flags; /* TT_FL_* */
+	int        flags; /* SCOLS_FL_* */
         const char *help;
 };
+
+/* basic output flags */
+static int no_headings;
+static int raw;
+
 enum { COL_PATH, COL_TYPE, COL_SIZE, COL_USED, COL_PRIO };
 struct colinfo infos[] = {
 	[COL_PATH]     = { "NAME",	0.20, 0, N_("device file or partition path") },
-	[COL_TYPE]     = { "TYPE",	0.20, TT_FL_TRUNC, N_("type of the device")},
-	[COL_SIZE]     = { "SIZE",	0.20, TT_FL_RIGHT, N_("size of the swap area")},
-	[COL_USED]     = { "USED",	0.20, TT_FL_RIGHT, N_("bytes in use")},
-	[COL_PRIO]     = { "PRIO",	0.20, TT_FL_RIGHT, N_("swap priority")},
+	[COL_TYPE]     = { "TYPE",	0.20, SCOLS_FL_TRUNC, N_("type of the device")},
+	[COL_SIZE]     = { "SIZE",	0.20, SCOLS_FL_RIGHT, N_("size of the swap area")},
+	[COL_USED]     = { "USED",	0.20, SCOLS_FL_RIGHT, N_("bytes in use")},
+	[COL_PRIO]     = { "PRIO",	0.20, SCOLS_FL_RIGHT, N_("swap priority")},
 };
-#define NCOLS ARRAY_SIZE(infos)
-static int columns[NCOLS], ncolumns;
+
+static int columns[ARRAY_SIZE(infos) * 2];
+static int ncolumns;
 
 static int column_name_to_id(const char *name, size_t namesz)
 {
@@ -101,7 +115,7 @@ static int column_name_to_id(const char *name, size_t namesz)
 
 	assert(name);
 
-	for (i = 0; i < NCOLS; i++) {
+	for (i = 0; i < ARRAY_SIZE(infos); i++) {
 		const char *cn = infos[i].name;
 
 		if (!strncasecmp(name, cn, namesz) && !*(cn + namesz))
@@ -113,9 +127,8 @@ static int column_name_to_id(const char *name, size_t namesz)
 
 static inline int get_column_id(int num)
 {
-	assert(ARRAY_SIZE(columns) == NCOLS);
 	assert(num < ncolumns);
-	assert(columns[num] < (int)NCOLS);
+	assert(columns[num] < (int) ARRAY_SIZE(infos));
 
 	return columns[num];
 }
@@ -125,37 +138,34 @@ static inline struct colinfo *get_column_info(unsigned num)
 	return &infos[get_column_id(num)];
 }
 
-static void add_tt_line(struct tt *tt, struct libmnt_fs *fs, int bytes)
+static void add_scols_line(struct libscols_table *table, struct libmnt_fs *fs, int bytes)
 {
 	int i;
-	struct tt_line *line;
+	struct libscols_line *line;
 
-	assert(tt);
+	assert(table);
 	assert(fs);
 
-	line = tt_add_line(tt, NULL);
-	if (!line) {
-		warn(_("failed to add line to output"));
-		return;
-	}
+	line = scols_table_new_line(table, NULL);
+	if (!line)
+		err(EXIT_FAILURE, _("failed to initialize output line"));
 
 	for (i = 0; i < ncolumns; i++) {
 		char *str = NULL;
-		int rc = 0;
 		off_t size;
 
 		switch (get_column_id(i)) {
 		case COL_PATH:
-			rc = xasprintf(&str, "%s", mnt_fs_get_source(fs));
+			xasprintf(&str, "%s", mnt_fs_get_source(fs));
 			break;
 		case COL_TYPE:
-			rc = xasprintf(&str, "%s", mnt_fs_get_swaptype(fs));
+			xasprintf(&str, "%s", mnt_fs_get_swaptype(fs));
 			break;
 		case COL_SIZE:
 			size = mnt_fs_get_size(fs);
 			size *= 1024;	/* convert to bytes */
 			if (bytes)
-				rc = xasprintf(&str, "%jd", size);
+				xasprintf(&str, "%jd", size);
 			else
 				str = size_to_human_string(SIZE_SUFFIX_1LETTER, size);
 			break;
@@ -163,19 +173,19 @@ static void add_tt_line(struct tt *tt, struct libmnt_fs *fs, int bytes)
 			size = mnt_fs_get_usedsize(fs);
 			size *= 1024;	/* convert to bytes */
 			if (bytes)
-				rc = xasprintf(&str, "%jd", size);
+				xasprintf(&str, "%jd", size);
 			else
 				str = size_to_human_string(SIZE_SUFFIX_1LETTER, size);
 			break;
 		case COL_PRIO:
-			rc = xasprintf(&str, "%d", mnt_fs_get_priority(fs));
+			xasprintf(&str, "%d", mnt_fs_get_priority(fs));
 			break;
 		default:
 			break;
 		}
 
-		if (rc || str)
-			tt_line_set_data(line, i, str);
+		if (str)
+			scols_line_refer_data(line, i, str);
 	}
 	return;
 }
@@ -189,15 +199,17 @@ static int display_summary(void)
 	if (!st)
 		return -1;
 
+	if (mnt_table_is_empty(st))
+		return 0;
+
 	itr = mnt_new_iter(MNT_ITER_FORWARD);
 	if (!itr)
 		err(EXIT_FAILURE, _("failed to initialize libmount iterator"));
 
-	if (mnt_table_get_nents(st) > 0)
-		printf(_("%s\t\t\t\tType\t\tSize\tUsed\tPriority\n"), _("Filename"));
+	printf(_("%s\t\t\t\tType\t\tSize\tUsed\tPriority\n"), _("Filename"));
 
 	while (mnt_table_next_fs(st, itr, &fs) == 0) {
-		printf("%-39s\t%s\t%jd\t%jd\t%d\n",
+		printf("%-39s\t%-8s\t%jd\t%jd\t%d\n",
 			mnt_fs_get_source(fs),
 			mnt_fs_get_swaptype(fs),
 			mnt_fs_get_size(fs),
@@ -209,14 +221,13 @@ static int display_summary(void)
 	return 0;
 }
 
-static int show_table(int tt_flags, int bytes)
+static int show_table(int bytes)
 {
 	struct libmnt_table *st = get_swaps();
 	struct libmnt_iter *itr = NULL;
 	struct libmnt_fs *fs;
-
-	int i, rc = 0;
-	struct tt *tt = NULL;
+	int i;
+	struct libscols_table *table = NULL;
 
 	if (!st)
 		return -1;
@@ -225,30 +236,29 @@ static int show_table(int tt_flags, int bytes)
 	if (!itr)
 		err(EXIT_FAILURE, _("failed to initialize libmount iterator"));
 
-	tt = tt_new_table(tt_flags);
-	if (!tt) {
-		warn(_("failed to initialize output table"));
-		goto done;
-	}
+	scols_init_debug(0);
+
+	table = scols_new_table();
+	if (!table)
+		err(EXIT_FAILURE, _("failed to initialize output table"));
+
+	scols_table_enable_raw(table, raw);
+	scols_table_enable_noheadings(table, no_headings);
 
 	for (i = 0; i < ncolumns; i++) {
 		struct colinfo *col = get_column_info(i);
 
-		if (!tt_define_column(tt, col->name, col->whint, col->flags)) {
-			warnx(_("failed to initialize output column"));
-			rc = -1;
-			goto done;
-		}
+		if (!scols_table_new_column(table, col->name, col->whint, col->flags))
+			err(EXIT_FAILURE, _("failed to initialize output column"));
 	}
 
 	while (mnt_table_next_fs(st, itr, &fs) == 0)
-		add_tt_line(tt, fs, bytes);
+		add_scols_line(table, fs, bytes);
 
-	tt_print_table(tt);
- done:
+	scols_print_table(table);
+	scols_unref_table(table);
 	mnt_free_iter(itr);
-	tt_free_table(tt);
-	return rc;
+	return 0;
 }
 
 /* calls mkswap */
@@ -280,7 +290,7 @@ static int swap_reinitialize(const char *device,
 		cmd[idx++] = (char *) device;
 		cmd[idx++] = NULL;
 		execv(cmd[0], cmd);
-		err(EXIT_FAILURE, _("execv failed"));
+		err(EXIT_FAILURE, _("failed to execute %s"), cmd[0]);
 
 	default: /* parent */
 		do {
@@ -324,14 +334,16 @@ static int swap_rewrite_signature(const char *devname, unsigned int pagesize)
 
 	rc  = 0;
 err:
-	close(fd);
+	if (close_fd(fd) != 0) {
+		warn(_("write failed: %s"), devname);
+		rc = -1;
+	}
 	return rc;
 }
 
 static int swap_detect_signature(const char *buf, int *sig)
 {
-	if (memcmp(buf, "SWAP-SPACE", 10) == 0 ||
-            memcmp(buf, "SWAPSPACE2", 10) == 0)
+	if (memcmp(buf, SWAP_SIGNATURE, SWAP_SIGNATURE_SZ) == 0)
 		*sig = SIG_SWAPSPACE;
 
 	else if (memcmp(buf, "S1SUSPEND", 9) == 0 ||
@@ -388,21 +400,19 @@ static unsigned long long swap_get_size(const char *hdr, const char *devname,
 					unsigned int pagesize)
 {
 	unsigned int last_page = 0;
-	int swap_version = 0;
+	const unsigned int swap_version = SWAP_VERSION;
 	int flip = 0;
 	struct swap_header_v1_2 *s;
 
 	s = (struct swap_header_v1_2 *) hdr;
-	if (s->version == 1) {
-		swap_version = 1;
+	if (s->version == swap_version) {
 		last_page = s->last_page;
-	} else if (swab32(s->version) == 1) {
+	} else if (swab32(s->version) == swap_version) {
 		flip = 1;
-		swap_version = 1;
 		last_page = swab32(s->last_page);
 	}
 	if (verbose)
-		warnx(_("%s: found swap signature: version %d, "
+		warnx(_("%s: found swap signature: version %ud, "
 			"page-size %d, %s byte order"),
 			devname,
 			swap_version,
@@ -441,26 +451,22 @@ static int swapon_checks(const char *special)
 	char *hdr = NULL;
 	unsigned int pagesize;
 	unsigned long long devsize = 0;
+	int permMask;
 
 	if (stat(special, &st) < 0) {
 		warn(_("stat failed %s"), special);
 		goto err;
 	}
 
-	/* people generally dislike this warning - now it is printed
-	   only when `verbose' is set */
-	if (verbose) {
-		int permMask = (S_ISBLK(st.st_mode) ? 07007 : 07077);
-
-		if ((st.st_mode & permMask) != 0)
-			warnx(_("%s: insecure permissions %04o, %04o suggested."),
+	permMask = S_ISBLK(st.st_mode) ? 07007 : 07077;
+	if ((st.st_mode & permMask) != 0)
+		warnx(_("%s: insecure permissions %04o, %04o suggested."),
 				special, st.st_mode & 07777,
 				~permMask & 0666);
 
-		if (S_ISREG(st.st_mode) && st.st_uid != 0)
-			warnx(_("%s: insecure file owner %d, 0 (root) suggested."),
+	if (S_ISREG(st.st_mode) && st.st_uid != 0)
+		warnx(_("%s: insecure file owner %d, 0 (root) suggested."),
 				special, st.st_uid);
-	}
 
 	/* test for holes by LBT */
 	if (S_ISREG(st.st_mode)) {
@@ -572,8 +578,22 @@ static int do_swapon(const char *orig_special, int prio,
 			   << SWAP_FLAG_PRIO_SHIFT);
 	}
 #endif
-	if (fl_discard)
-		flags |= SWAP_FLAG_DISCARD;
+	/*
+	 * Validate the discard flags passed and set them
+	 * accordingly before calling sys_swapon.
+	 */
+	if (fl_discard && !(fl_discard & ~SWAP_FLAGS_DISCARD_VALID)) {
+		/*
+		 * If we get here with both discard policy flags set,
+		 * we just need to tell the kernel to enable discards
+		 * and it will do correctly, just as we expect.
+		 */
+		if ((fl_discard & SWAP_FLAG_DISCARD_ONCE) &&
+		    (fl_discard & SWAP_FLAG_DISCARD_PAGES))
+			flags |= SWAP_FLAG_DISCARD;
+		else
+			flags |= fl_discard;
+	}
 
 	status = swapon(special, flags);
 	if (status < 0)
@@ -613,12 +633,22 @@ static int swapon_all(void)
 	while (mnt_table_find_next_fs(tb, itr, match_swap, NULL, &fs) == 0) {
 		/* defaults */
 		int pri = priority, dsc = discard, nofail = ifexists;
-		char *p, *src;
+		char *p, *src, *dscarg;
 
 		if (mnt_fs_get_option(fs, "noauto", NULL, NULL) == 0)
 			continue;
-		if (mnt_fs_get_option(fs, "discard", NULL, NULL) == 0)
-			dsc = 1;
+		if (mnt_fs_get_option(fs, "discard", &dscarg, NULL) == 0) {
+			dsc |= SWAP_FLAG_DISCARD;
+			if (dscarg) {
+				/* only single-time discards are wanted */
+				if (strcmp(dscarg, "once") == 0)
+					dsc |= SWAP_FLAG_DISCARD_ONCE;
+
+				/* do discard for every released swap page */
+				if (strcmp(dscarg, "pages") == 0)
+					dsc |= SWAP_FLAG_DISCARD_PAGES;
+			}
+		}
 		if (mnt_fs_get_option(fs, "nofail", NULL, NULL) == 0)
 			nofail = 1;
 		if (mnt_fs_get_option(fs, "pri", &p, NULL) == 0 && p)
@@ -648,17 +678,17 @@ static void __attribute__ ((__noreturn__)) usage(FILE * out)
 	fprintf(out, _(" %s [options] [<spec>]\n"), program_invocation_short_name);
 
 	fputs(USAGE_OPTIONS, out);
-	fputs(_(" -a, --all              enable all swaps from /etc/fstab\n"
-		" -d, --discard          discard freed pages before they are reused\n"
-		" -e, --ifexists         silently skip devices that do not exist\n"
-		" -f, --fixpgsz          reinitialize the swap space if necessary\n"
-		" -p, --priority <prio>  specify the priority of the swap device\n"
-		" -s, --summary          display summary about used swap devices\n"
-		"     --show[=<columns>] display summary in definable table\n"
-		"     --noheadings       don't print headings, use with --show\n"
-		"     --raw              use the raw output format, use with --show\n"
-		"     --bytes            display swap size in bytes in --show output\n"
-		" -v, --verbose          verbose mode\n"), out);
+	fputs(_(" -a, --all                enable all swaps from /etc/fstab\n"
+		" -d, --discard[=<policy>] enable swap discards, if supported by device\n"
+		" -e, --ifexists           silently skip devices that do not exist\n"
+		" -f, --fixpgsz            reinitialize the swap space if necessary\n"
+		" -p, --priority <prio>    specify the priority of the swap device\n"
+		" -s, --summary            display summary about used swap devices (DEPRECATED)\n"
+		"     --show[=<columns>]   display summary in definable table\n"
+		"     --noheadings         don't print headings, use with --show\n"
+		"     --raw                use the raw output format, use with --show\n"
+		"     --bytes              display swap size in bytes in --show output\n"
+		" -v, --verbose            verbose mode\n"), out);
 
 	fputs(USAGE_SEPARATOR, out);
 	fputs(USAGE_HELP, out);
@@ -674,8 +704,13 @@ static void __attribute__ ((__noreturn__)) usage(FILE * out)
 		" <device>               name of device to be used\n"
 		" <file>                 name of file to be used\n"), out);
 
+	fputs(_("\nAvailable discard policy types (for --discard):\n"
+		" once	  : only single-time area discards are issued. (swapon)\n"
+		" pages	  : discard freed pages before they are reused.\n"
+		" * if no policy is selected both discard types are enabled. (default)\n"), out);
+
 	fputs(_("\nAvailable columns (for --show):\n"), out);
-	for (i = 0; i < NCOLS; i++)
+	for (i = 0; i < ARRAY_SIZE(infos); i++)
 		fprintf(out, " %4s  %s\n", infos[i].name, _(infos[i].help));
 
 	fprintf(out, USAGE_MAN_TAIL("swapon(8)"));
@@ -685,7 +720,7 @@ static void __attribute__ ((__noreturn__)) usage(FILE * out)
 int main(int argc, char *argv[])
 {
 	int status = 0, c;
-	int show = 0, tt_flags = 0;
+	int show = 0;
 	int bytes = 0;
 	size_t i;
 
@@ -698,7 +733,7 @@ int main(int argc, char *argv[])
 
 	static const struct option long_opts[] = {
 		{ "priority", 1, 0, 'p' },
-		{ "discard",  0, 0, 'd' },
+		{ "discard",  2, 0, 'd' },
 		{ "ifexists", 0, 0, 'e' },
 		{ "summary",  0, 0, 's' },
 		{ "fixpgsz",  0, 0, 'f' },
@@ -721,7 +756,7 @@ int main(int argc, char *argv[])
 	mnt_init_debug(0);
 	mntcache = mnt_new_cache();
 
-	while ((c = getopt_long(argc, argv, "ahdefp:svVL:U:",
+	while ((c = getopt_long(argc, argv, "ahd::efp:svVL:U:",
 				long_opts, NULL)) != -1) {
 		switch (c) {
 		case 'a':		/* all */
@@ -731,7 +766,8 @@ int main(int argc, char *argv[])
 			usage(stdout);
 			break;
 		case 'p':		/* priority */
-			priority = atoi(optarg);
+			priority = strtos16_or_err(optarg,
+					   _("failed to parse priority"));
 			break;
 		case 'L':
 			add_label(optarg);
@@ -740,7 +776,18 @@ int main(int argc, char *argv[])
 			add_uuid(optarg);
 			break;
 		case 'd':
-			discard = 1;
+			discard |= SWAP_FLAG_DISCARD;
+			if (optarg) {
+				if (*optarg == '=')
+					optarg++;
+
+				if (strcmp(optarg, "once") == 0)
+					discard |= SWAP_FLAG_DISCARD_ONCE;
+				else if (strcmp(optarg, "pages") == 0)
+					discard |= SWAP_FLAG_DISCARD_PAGES;
+				else
+					errx(EXIT_FAILURE, _("unsupported discard policy: %s"), optarg);
+			}
 			break;
 		case 'e':               /* ifexists */
 		        ifexists = 1;
@@ -766,10 +813,10 @@ int main(int argc, char *argv[])
 			show = 1;
 			break;
 		case NOHEADINGS_OPTION:
-			tt_flags |= TT_FL_NOHEADINGS;
+			no_headings = 1;
 			break;
 		case RAW_OPTION:
-			tt_flags |= TT_FL_RAW;
+			raw = 1;
 			break;
 		case BYTES_OPTION:
 			bytes = 1;
@@ -786,7 +833,7 @@ int main(int argc, char *argv[])
 	}
 	argv += optind;
 
-	if (show) {
+	if (show || (!all && !numof_labels() && !numof_uuids() && *argv == NULL)) {
 		if (!ncolumns) {
 			/* default columns */
 			columns[ncolumns++] = COL_PATH;
@@ -795,13 +842,10 @@ int main(int argc, char *argv[])
 			columns[ncolumns++] = COL_USED;
 			columns[ncolumns++] = COL_PRIO;
 		}
-		status = show_table(tt_flags, bytes);
+		status = show_table(bytes);
 		return status;
 	}
 
-	if (!all && !numof_labels() && !numof_uuids() && *argv == NULL)
-		usage(stderr);
-
 	if (ifexists && !all)
 		usage(stderr);
 
@@ -818,7 +862,7 @@ int main(int argc, char *argv[])
 		status |= do_swapon(*argv++, priority, discard, !CANONIC);
 
 	free_tables();
-	mnt_free_cache(mntcache);
+	mnt_unref_cache(mntcache);
 
 	return status;
 }
